@@ -4,13 +4,13 @@
 #
 ################################################################################
 #
-#  $Revision: 11 $
+#  $Revision: 19 $
 #  $Author: mhx $
-#  $Date: 2004/08/13 12:50:05 +0200 $
+#  $Date: 2007/08/13 22:59:58 +0200 $
 #
 ################################################################################
 #
-#  Version 3.x, Copyright (C) 2004, Marcus Holland-Moritz.
+#  Version 3.x, Copyright (C) 2004-2007, Marcus Holland-Moritz.
 #  Version 2.x, Copyright (C) 2001, Paul Marquess.
 #  Version 1.x, Copyright (C) 1999, Kenneth Albanowski.
 #
@@ -19,6 +19,24 @@
 #
 ################################################################################
 
+sub cat_file
+{
+  eval { require File::Spec };
+  return $@ ? join('/', @_) : File::Spec->catfile(@_);
+}
+
+sub all_files_in_dir
+{
+  my $dir = shift;
+  local *DIR;
+
+  opendir DIR, $dir or die "cannot open directory $dir: $!\n";
+  my @files = grep { !-d && !/^\./ } readdir DIR;  # no dirs or hidden files
+  closedir DIR;
+
+  return map { cat_file($dir, $_) } @files;
+}
+
 sub parse_todo
 {
   my $dir = shift || 'parts/todo';
@@ -26,7 +44,7 @@ sub parse_todo
   my %todo;
   my $todo;
 
-  for $todo (glob "$dir/*") {
+  for $todo (all_files_in_dir($dir)) {
     open TODO, $todo or die "cannot open $todo: $!\n";
     my $perl = <TODO>;
     chomp $perl;
@@ -45,6 +63,17 @@ sub parse_todo
   return \%todo;
 }
 
+sub expand_version
+{
+  my($op, $ver) = @_;
+  my($r, $v, $s) = parse_version($ver);
+  $r == 5 or die "only Perl revision 5 is supported\n";
+  $op eq '=='     and return "((PERL_VERSION == $v) && (PERL_SUBVERSION == $s))";
+  $op eq '!='     and return "((PERL_VERSION != $v) || (PERL_SUBVERSION != $s))";
+  $op =~ /([<>])/ and return "((PERL_VERSION $1 $v) || ((PERL_VERSION == $v) && (PERL_SUBVERSION $op $s)))";
+  die "cannot expand version expression ($op $ver)\n";
+}
+
 sub parse_partspec
 {
   my $file = shift;
@@ -139,6 +168,12 @@ sub parse_partspec
     }
   }
 
+  for $section (qw( implementation xsubs xsinit xsmisc xshead xsboot )) {
+    if (exists $data{$section}) {
+      $data{$section} =~ s/\{\s*version\s*(<|>|==|!=|>=|<=)\s*([\d._]+)\s*\}/expand_version($1, $2)/gei;
+    }
+  }
+
   $data{provides}   = \@prov;
   $data{prototypes} = \%proto;
   $data{OPTIONS}    = \%options;
@@ -154,7 +189,7 @@ sub parse_partspec
                              (defined $nop && exists $dontwarn{$nop}) ||
                              $h{$_}++;
                        }
-                       $data{implementation} =~ /^\s*#\s*define\s+(\w+)/g };
+                       $data{implementation} =~ /^\s*#\s*define\s+(\w+)/gm };
 
   if (@maybeprov) {
     warn "$file seems to provide these macros, but doesn't list them:\n  "
@@ -194,15 +229,17 @@ sub ppcond
 sub trim_arg
 {
   my $in = shift;
+  my $remove = join '|', qw( NN NULLOK );
 
   $in eq '...' and return ($in);
 
   local $_ = $in;
   my $id;
- 
+
   s/[*()]/ /g;
   s/\[[^\]]*\]/ /g;
   s/\b(?:auto|const|extern|inline|register|static|volatile|restrict)\b//g;
+  s/\b(?:$remove)\b//;
   s/^\s*//; s/\s*$//;
 
   if( /^\b(?:struct|union|enum)\s+\w+(?:\s+(\w+))?$/ ) {
@@ -223,6 +260,7 @@ sub trim_arg
 
   # these don't matter at all
   s/\b(?:auto|extern|inline|register|static|volatile|restrict)\b//g;
+  s/\b(?:$remove)\b//;
 
   s/(?=<\*)\s+(?=\*)//g;
   s/\s*(\*+)\s*/ $1 /g;
